import unittest
from unittest import mock

from flumine import Flumine, worker
from flumine.events import events
from flumine.clients import ExchangeType


class FlumineTest(unittest.TestCase):
    def setUp(self):
        self.mock_client = mock.Mock(EXCHANGE=ExchangeType.SIMULATED)
        self.flumine = Flumine(self.mock_client)

    @mock.patch("flumine.flumine.Flumine._add_default_workers")
    @mock.patch("flumine.flumine.Flumine._process_custom_event")
    @mock.patch("flumine.flumine.Flumine._process_cleared_orders")
    @mock.patch("flumine.flumine.Flumine._process_cleared_markets")
    @mock.patch("flumine.flumine.Flumine._process_close_market")
    @mock.patch("flumine.flumine.Flumine._process_current_orders")
    @mock.patch("flumine.flumine.Flumine._process_end_flumine")
    @mock.patch("flumine.flumine.Flumine._process_market_catalogues")
    @mock.patch("flumine.flumine.Flumine._process_raw_data")
    @mock.patch("flumine.flumine.Flumine._process_sports_data")
    @mock.patch("flumine.flumine.Flumine._process_market_books")
    def test_run(
        self,
        mock__process_market_books,
        mock__process_sports_data,
        mock__process_raw_data,
        mock__process_market_catalogues,
        mock__process_end_flumine,
        mock__process_current_orders,
        mock__process_close_market,
        mock__process_cleared_markets,
        mock__process_cleared_orders,
        mock__process_custom_event,
        mock__add_default_workers,
    ):
        mock_events = [
            events.MarketCatalogueEvent(None),
            events.MarketBookEvent(None),
            events.SportsDataEvent(None),
            events.RawDataEvent(None),
            events.CurrentOrdersEvent(None),
            events.ClearedMarketsEvent(None),
            events.ClearedOrdersEvent(None),
            events.CloseMarketEvent(None),
            events.CustomEvent(None, None),
            events.TerminationEvent(None),
        ]
        for i in mock_events:
            self.flumine.handler_queue.put(i)
        self.flumine.run()

        mock__process_market_books.assert_called_with(mock_events[1])
        mock__process_sports_data.assert_called_with(mock_events[2])
        mock__process_raw_data.assert_called_with(mock_events[3])
        mock__process_market_catalogues.assert_called_with(mock_events[0])
        mock__process_end_flumine.assert_called_with()
        mock__process_current_orders.assert_called_with(mock_events[4])
        mock__process_close_market.assert_called_with(mock_events[7])
        mock__process_cleared_markets.assert_called_with(mock_events[5])
        mock__process_cleared_orders.assert_called_with(mock_events[6])
        mock__process_custom_event.assert_called_with(mock_events[8])
        mock__add_default_workers.assert_called()

    @mock.patch("flumine.worker.BackgroundWorker")
    @mock.patch("flumine.Flumine.add_worker")
    def test__add_default_workers(self, mock_add_worker, mock_worker):
        mock_client_one = mock.Mock(market_recording_mode=False)
        mock_client_one.betting_client.session_timeout = 1200
        mock_client_two = mock.Mock(market_recording_mode=True)
        mock_client_two.betting_client.session_timeout = 600
        self.flumine.clients = [
            mock_client_one,
            mock_client_two,
        ]
        self.flumine._add_default_workers()
        self.assertEqual(
            mock_worker.call_args_list,
            [
                mock.call(self.flumine, function=worker.keep_alive, interval=300),
                mock.call(
                    self.flumine,
                    function=worker.poll_market_catalogue,
                    interval=60,
                    start_delay=10,
                ),
                mock.call(
                    self.flumine,
                    function=worker.poll_account_balance,
                    interval=120,
                    start_delay=10,
                ),
                mock.call(
                    self.flumine,
                    function=worker.poll_market_closure,
                    interval=60,
                    start_delay=10,
                ),
            ],
        )

    @mock.patch("flumine.worker.BackgroundWorker")
    @mock.patch("flumine.Flumine.add_worker")
    def test__add_default_workers_market_record(self, mock_add_worker, mock_worker):
        mock_client_one = mock.Mock(market_recording_mode=True)
        mock_client_one.betting_client.session_timeout = 1200
        mock_client_two = mock.Mock(market_recording_mode=True)
        mock_client_two.betting_client.session_timeout = 600
        self.flumine.clients = [
            mock_client_one,
            mock_client_two,
        ]
        self.flumine._add_default_workers()
        self.assertEqual(
            mock_worker.call_args_list,
            [
                mock.call(self.flumine, function=worker.keep_alive, interval=300),
                mock.call(
                    self.flumine,
                    function=worker.poll_market_catalogue,
                    interval=60,
                    start_delay=10,
                ),
            ],
        )

    def test_str(self):
        assert str(self.flumine) == "<Flumine>"

    def test_repr(self):
        assert repr(self.flumine) == "<Flumine>"



import unittest
import datetime
from unittest import mock
from unittest.mock import call

from flumine.streams import streams, datastream, historicalstream
from flumine.streams.basestream import BaseStream
from flumine.streams.simulatedorderstream import CurrentOrders
from flumine.streams import orderstream
from flumine.exceptions import ListenerError


class StreamsTest(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_flumine = mock.Mock()
        self.mock_flumine.SIMULATED = False
        self.streams = streams.Streams(self.mock_flumine)

    def test_init(self):
        self.assertEqual(self.streams.flumine, self.mock_flumine)
        self.assertEqual(self.streams._streams, [])
        self.assertEqual(self.streams._stream_id, 0)

    @mock.patch("flumine.streams.streams.Streams.add_stream")
    def test_call(self, mock_add_stream):
        mock_strategy = mock.Mock(streams=[], raw_data=False)
        self.streams(mock_strategy)
        mock_add_stream.assert_called_with(mock_strategy)

    @mock.patch("flumine.streams.streams.Streams.add_stream")
    def test_call_data_stream(self, mock_add_stream):
        mock_strategy = mock.Mock(streams=[], stream_class=datastream.DataStream)
        self.streams(mock_strategy)
        mock_add_stream.assert_called_with(mock_strategy)

    @mock.patch("flumine.streams.streams.get_file_md")
    @mock.patch("flumine.streams.streams.Streams.add_historical_stream")
    def test_call_simulated_markets(self, mock_add_historical_stream, mock_get_file_md):
        self.mock_flumine.SIMULATED = True
        mock_md = mock.Mock()
        mock_get_file_md.return_value = mock_md

        mock_strategy = mock.Mock(
            streams=[],
            historic_stream_ids=set(),
            market_filter={
                "markets": ["dubs of the mad skint and british"],
                "listener_kwargs": {"canary_yellow": True},
            },
        )
        self.streams(mock_strategy)

        mock_add_historical_stream.assert_called_with(
            mock_strategy,
            "dubs of the mad skint and british",
            mock_md,
            False,
            canary_yellow=True,
        )
        self.assertEqual(len(mock_strategy.streams), 1)
        self.assertEqual(len(mock_strategy.historic_stream_ids), 1)
        mock_get_file_md.assert_has_calls(
            [
                call("dubs of the mad skint and british"),
            ]
        )

    @mock.patch("flumine.streams.streams.get_file_md")
    @mock.patch("flumine.streams.streams.Streams.add_historical_stream")
    def test_call_simulated_markets_type(
        self, mock_add_historical_stream, mock_get_file_md
    ):
        self.mock_flumine.SIMULATED = True
        mock_get_file_md.return_value = mock.Mock(market_type="PLACE")
        mock_strategy = mock.Mock(
            streams=[],
            historic_stream_ids=[],
            market_filter={
                "markets": ["dubs of the mad skint and british"],
                "listener_kwargs": {"canary_yellow": True},
                "market_types": ["WIN"],
            },
        )
        self.streams(mock_strategy)

        mock_add_historical_stream.assert_not_called()
        self.assertEqual(len(mock_strategy.streams), 0)
        self.assertEqual(len(mock_strategy.historic_stream_ids), 0)
        mock_get_file_md.assert_has_calls(
            [
                call("dubs of the mad skint and british"),
            ]
        )

    @mock.patch("flumine.streams.streams.get_file_md")
    @mock.patch("flumine.streams.streams.Streams.add_historical_stream")
    def test_call_simulated_country_code(
        self, mock_add_historical_stream, mock_get_file_md
    ):
        self.mock_flumine.SIMULATED = True
        mock_get_file_md.return_value = mock.Mock(market_type="PLACE")
        mock_strategy = mock.Mock(
            streams=[],
            historic_stream_ids=[],
            market_filter={
                "markets": ["dubs of the mad skint and british"],
                "listener_kwargs": {"canary_yellow": True},
                "country_codes": ["VERDANSK"],
            },
        )
        self.streams(mock_strategy)

        mock_add_historical_stream.assert_not_called()
        self.assertEqual(len(mock_strategy.streams), 0)
        self.assertEqual(len(mock_strategy.historic_stream_ids), 0)
        mock_get_file_md.assert_has_calls(
            [
                call("dubs of the mad skint and british"),
            ]
        )

    @mock.patch("flumine.streams.streams.Streams.add_historical_stream")
    def test_call_simulated_events(self, mock_add_historical_stream):
        self.mock_flumine.SIMULATED = True
        mock_strategy = mock.Mock(
            streams=[],
            historic_stream_ids=[],
            market_filter={
                "events": ["/tmp/joetry"],
                "listener_kwargs": {"canary_yellow": True},
            },
        )
        with self.assertRaises(NotImplementedError):
            self.streams(mock_strategy)

        # mock_add_historical_stream.assert_called_with(
        #     mock_strategy, "dubs of the mad skint and british", canary_yellow=True
        # )
        # self.assertEqual(len(mock_strategy.streams), 1)
        # self.assertEqual(len(mock_strategy.historic_stream_ids), 1)

    def test_call_simulated_markets_events(self):
        self.mock_flumine.SIMULATED = True
        mock_strategy = mock.Mock(
            streams=[],
            market_filter={
                "markets": ["dubs of the mad skint and british"],
                "events": ["joetry"],
            },
        )
        self.streams(mock_strategy)
        self.assertEqual(len(mock_strategy.streams), 0)

    def test_call_simulated_no_markets_no_events(self):
        self.mock_flumine.SIMULATED = True
        mock_strategy = mock.Mock(streams=[], market_filter={})
        self.streams(mock_strategy)
        self.assertEqual(len(mock_strategy.streams), 0)

    @mock.patch("flumine.streams.streams.Streams.add_order_stream")
    def test_add_client_betfair(self, mock_add_order_stream):
        mock_client = mock.Mock(order_stream=True, paper_trade=False)
        mock_client.EXCHANGE = streams.ExchangeType.BETFAIR
        self.streams.add_client(mock_client)
        mock_add_order_stream.assert_called_with(mock_client)

    @mock.patch("flumine.streams.streams.Streams.add_simulated_order_stream")
    def test_add_client_paper_trade(self, mock_add_simulated_order_stream):
        mock_client = mock.Mock(order_stream=True, paper_trade=True)
        mock_client.EXCHANGE = streams.ExchangeType.BETFAIR
        self.streams.add_client(mock_client)
        mock_add_simulated_order_stream.assert_called_with(mock_client)

    @mock.patch("flumine.streams.streams.Streams.add_order_stream")
    def test_add_client_no_order_stream(self, mock_add_order_stream):
        mock_client = mock.Mock(order_stream=False)
        mock_client.EXCHANGE = streams.ExchangeType.BETFAIR
        self.streams.add_client(mock_client)
        mock_add_order_stream.assert_not_called()

    @mock.patch("flumine.streams.streams.Streams._increment_stream_id")
    def test_add_stream_new(self, mock_increment):
        mock_strategy = mock.Mock(market_filter={}, sports_data_filter=[])
        mock_stream_class = mock.Mock()
        mock_strategy.stream_class = mock_stream_class

        self.streams.add_stream(mock_strategy)
        self.assertEqual(len(self.streams), 1)
        mock_increment.assert_called_with()
        mock_strategy.stream_class.assert_called_with(
            flumine=self.mock_flumine,
            stream_id=mock_increment(),
            market_filter=mock_strategy.market_filter,
            market_data_filter=mock_strategy.market_data_filter,
            streaming_timeout=mock_strategy.streaming_timeout,
            conflate_ms=mock_strategy.conflate_ms,
        )

    @mock.patch("flumine.streams.streams.Streams._increment_stream_id")
    def test_add_stream_none(self, mock_increment):
        mock_strategy = mock.Mock(market_filter=None, sports_data_filter=[])
        mock_stream_class = mock.Mock()
        mock_strategy.stream_class = mock_stream_class

        self.streams.add_stream(mock_strategy)
        self.assertEqual(len(self.streams), 1)
        mock_increment.assert_called_with()
        mock_strategy.stream_class.assert_called_with(
            flumine=self.mock_flumine,
            stream_id=mock_increment(),
            market_filter=mock_strategy.market_filter,
            market_data_filter=mock_strategy.market_data_filter,
            streaming_timeout=mock_strategy.streaming_timeout,
            conflate_ms=mock_strategy.conflate_ms,
        )

    @mock.patch("flumine.streams.streams.Streams._increment_stream_id")
    def test_add_stream_multi(self, mock_increment):
        mock_strategy = mock.Mock(
            market_filter=[{1: 2}, {3: 4}],
            stream_class=streams.MarketStream,
            sports_data_filter=[],
        )
        self.streams.add_stream(mock_strategy)
        self.assertEqual(len(self.streams), 2)
        mock_increment.assert_called_with()

    @mock.patch("flumine.streams.streams.Streams._increment_stream_id")
    def test_add_stream_old(self, mock_increment):
        mock_strategy = mock.Mock(
            market_data_filter=2,
            stream_class=streams.MarketStream,
            streaming_timeout=3,
            conflate_ms=4,
            market_filter={},
            sports_data_filter=[],
        )
        stream = mock.Mock(
            spec=streams.MarketStream,
            market_filter={},
            market_data_filter=2,
            streaming_timeout=3,
            conflate_ms=4,
            stream_id=1001,
        )
        self.streams._streams = [stream]

        self.streams.add_stream(mock_strategy)
        self.assertEqual(len(self.streams), 1)
        mock_increment.assert_not_called()

    @mock.patch(
        "flumine.streams.streams.SportsDataStream", autospec=streams.SportsDataStream
    )
    @mock.patch("flumine.streams.streams.Streams._increment_stream_id")
    def test_add_stream_sports_data(self, mock_increment, sports_data_stream_class):
        mock_strategy = mock.Mock(
            market_filter=[],
            sports_data_filter=["raceSubscription"],
        )
        self.streams.add_stream(mock_strategy)
        self.assertEqual(len(self.streams), 1)
        mock_increment.assert_called_with()
        sports_data_stream_class.assert_has_calls(
            [
                call(
                    flumine=self.mock_flumine,
                    stream_id=mock_increment(),
                    sports_data_filter="raceSubscription",
                    streaming_timeout=mock_strategy.streaming_timeout,
                )
            ]
        )

    @mock.patch("flumine.streams.streams.Streams._increment_stream_id")
    def test_add_stream_sports_data_old(self, mock_increment):
        mock_strategy = mock.Mock(
            market_filter=[],
            sports_data_filter=["cricketSubscription"],
            streaming_timeout=3,
        )
        stream = mock.Mock(
            stream_id=1,
            spec=streams.SportsDataStream,
            market_filter={},
            sports_data_filter="cricketSubscription",
            streaming_timeout=3,
        )
        self.streams._streams = [stream]

        self.streams.add_stream(mock_strategy)
        self.assertEqual(len(self.streams), 1)
        mock_increment.assert_not_called()

    @mock.patch("flumine.streams.streams.HistoricalStream")
    @mock.patch("flumine.streams.streams.Streams._increment_stream_id")
    def test_add_historical_stream(self, mock_increment, mock_historical_stream_class):
        mock_historical_stream_class.__name__ = "test"
        self.mock_flumine.SIMULATED = True
        mock_strategy = mock.Mock()
        mock_strategy.market_filter = 1
        mock_strategy.market_data_filter = 2
        mock_strategy.streaming_timeout = 3
        mock_strategy.conflate_ms = 4
        mock_strategy.stream_class = streams.MarketStream
        mock_event_id = "12345"
        mock_md = mock.Mock(event_id=mock_event_id)

        self.streams.add_historical_stream(
            mock_strategy, "GANG", mock_md, event_processing=False, inplay=True
        )
        self.assertEqual(len(self.streams), 1)
        mock_increment.assert_called_with()
        mock_historical_stream_class.assert_called_with(
            flumine=self.mock_flumine,
            stream_id=mock_increment(),
            market_filter="GANG",
            market_data_filter=mock_strategy.market_data_filter,
            streaming_timeout=mock_strategy.streaming_timeout,
            conflate_ms=mock_strategy.conflate_ms,
            output_queue=False,
            event_processing=False,
            event_id=mock_event_id,
            inplay=True,
        )

    def test_add_historical_stream_old(self):
        self.mock_flumine.SIMULATED = True
        mock_strategy = mock.Mock()
        mock_stream = mock.Mock(event_processing=False, listener_kwargs={})
        mock_stream.market_filter = "GANG"
        self.streams._streams = [mock_stream]
        mock_md = mock.Mock()

        stream = self.streams.add_historical_stream(
            mock_strategy, "GANG", mock_md, event_processing=False, **{}
        )
        self.assertEqual(stream, mock_stream)
        self.assertEqual(len(self.streams), 1)

    @mock.patch("flumine.streams.streams.HistoricalStream")
    @mock.patch("flumine.streams.streams.Streams._increment_stream_id")
    def test_add_historical_stream_kwargs(
        self, mock_increment, mock_historical_stream_class
    ):
        mock_historical_stream_class.__name__ = "test"
        self.mock_flumine.SIMULATED = True
        mock_strategy = mock.Mock()
        mock_stream = mock.Mock(event_processing=False, listener_kwargs={})
        mock_stream.market_filter = "GANG"
        self.streams._streams = [mock_stream]
        mock_event_id = "12345"
        mock_md = mock.Mock(event_id=mock_event_id)

        self.streams.add_historical_stream(
            mock_strategy, "GANG", mock_md, event_processing=False, **{"inplay": True}
        )
        self.assertEqual(len(self.streams), 2)
        mock_increment.assert_called_with()
        mock_historical_stream_class.assert_called_with(
            flumine=self.mock_flumine,
            stream_id=mock_increment(),
            market_filter="GANG",
            market_data_filter=mock_strategy.market_data_filter,
            streaming_timeout=mock_strategy.streaming_timeout,
            conflate_ms=mock_strategy.conflate_ms,
            output_queue=False,
            event_processing=False,
            event_id=mock_event_id,
            inplay=True,
        )

    @mock.patch("flumine.streams.streams.HistoricalStream")
    def test_add_historical_stream_event_processing(self, mock_historical_stream_class):
        mock_historical_stream_class.__name__ = "test"
        self.mock_flumine.SIMULATED = True
        mock_strategy = mock.Mock()
        mock_stream = mock.Mock(event_processing=False)
        mock_stream.market_filter = "GANG"
        self.streams._streams = [mock_stream]
        mock_md = mock.Mock()

        stream = self.streams.add_historical_stream(
            mock_strategy, "GANG", mock_md, event_processing=True
        )
        self.assertEqual(stream, mock_historical_stream_class())
        self.assertEqual(len(self.streams), 2)

    @mock.patch("flumine.streams.streams.OrderStream")
    @mock.patch("flumine.streams.streams.Streams._increment_stream_id")
    def test_add_historical_stream_conflate(
        self, mock_increment, mock_order_stream_class
    ):
        conflate_ms = 500
        streaming_timeout = 0.5
        mock_client = mock.Mock()
        self.streams.add_order_stream(mock_client, conflate_ms, streaming_timeout)
        self.assertEqual(len(self.streams), 1)
        mock_increment.assert_called_with()
        mock_order_stream_class.assert_called_with(
            flumine=self.mock_flumine,
            stream_id=mock_increment(),
            streaming_timeout=streaming_timeout,
            conflate_ms=conflate_ms,
            client=mock_client,
        )

    @mock.patch("flumine.streams.streams.SimulatedOrderStream")
    @mock.patch("flumine.streams.streams.Streams._increment_stream_id")
    def test_add_simulated_order_stream(self, mock_increment, mock_order_stream_class):
        conflate_ms = 500
        streaming_timeout = 0.5
        mock_client = mock.Mock()
        self.streams.add_simulated_order_stream(
            mock_client, conflate_ms, streaming_timeout
        )
        self.assertEqual(len(self.streams), 1)
        mock_increment.assert_called_with()
        mock_order_stream_class.assert_called_with(
            flumine=self.mock_flumine,
            stream_id=mock_increment(),
            streaming_timeout=streaming_timeout,
            conflate_ms=conflate_ms,
            client=mock_client,
            custom=True,
        )

    @mock.patch("flumine.streams.streams.Streams._increment_stream_id")
    def test_add_custom_stream(self, mock_increment):
        mock_stream = mock.Mock()
        self.streams.add_custom_stream(mock_stream)
        mock_increment.assert_called_with()
        self.assertEqual(self.streams._streams, [mock_stream])

    def test_start(self):
        mock_stream = mock.Mock()
        self.streams._streams = [mock_stream]
        self.streams.start()
        mock_stream.start.assert_called_with()

    def test_start_simulated(self):
        self.mock_flumine.SIMULATED = True
        mock_stream = mock.Mock()
        self.streams._streams = [mock_stream]
        self.streams.start()
        mock_stream.start.assert_not_called()

    def test_stop(self):
        mock_stream = mock.Mock()
        self.streams._streams = [mock_stream]
        self.streams.stop()
        mock_stream.stop.assert_called_with()

    def test__increment_stream_id(self):
        self.assertEqual(self.streams._increment_stream_id(), 10000)

    def test_iter(self):
        for i in self.streams:
            assert i

    def test_len(self):
        self.assertEqual(len(self.streams), 0)


class TestBaseStream(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_flumine = mock.Mock()
        self.mock_client = mock.Mock()
        self.stream = BaseStream(
            self.mock_flumine,
            123,
            0.01,
            100,
            {"test": "me"},
            {"please": "now"},
            "yes",
            client=self.mock_client,
            output_queue=False,
            event_processing=True,
            event_id="123",
            operation="test",
            **{"calculate_market_tv": True},
        )

    def test_init(self):
        self.assertEqual(self.stream.flumine, self.mock_flumine)
        self.assertEqual(self.stream.stream_id, 123)
        self.assertEqual(self.stream.market_filter, {"test": "me"})
        self.assertEqual(self.stream.market_data_filter, {"please": "now"})
        self.assertEqual(self.stream.sports_data_filter, "yes")
        self.assertEqual(self.stream.streaming_timeout, 0.01)
        self.assertEqual(self.stream.conflate_ms, 100)
        self.assertFalse(self.stream.custom)
        self.assertIsNone(self.stream._stream)
        self.assertEqual(self.stream._client, self.mock_client)
        self.assertEqual(self.stream.MAX_LATENCY, 0.5)
        self.assertIsNone(self.stream._output_queue)
        self.assertTrue(self.stream.event_processing)
        self.assertEqual(self.stream.event_id, "123")
        self.assertEqual(self.stream.operation, "test")
        self.assertEqual(self.stream.listener_kwargs, {"calculate_market_tv": True})

    def test_run(self):
        with self.assertRaises(NotImplementedError):
            self.stream.run()

    def test_handle_output(self):
        with self.assertRaises(NotImplementedError):
            self.stream.handle_output()

    def test_stop(self):
        mock_stream = mock.Mock()
        self.stream._stream = mock_stream
        self.stream.stop()
        mock_stream.stop.assert_called_with()

    @mock.patch("flumine.streams.basestream.BaseStream.client")
    def test_betting_client(self, mock_client):
        self.assertEqual(self.stream.betting_client, mock_client.betting_client)

    def test_client(self):
        self.stream._client = 1
        self.assertEqual(self.stream.client, 1)
        self.stream._client = None
        self.assertEqual(self.stream.client, self.mock_flumine.clients.get_default())

    def test_stream_running(self):
        self.assertFalse(self.stream.stream_running)
        mock_stream = mock.Mock(running=True)
        self.stream._stream = mock_stream
        self.assertTrue(self.stream.stream_running)


class TestMarketStream(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_flumine = mock.Mock()
        self.stream = streams.MarketStream(
            self.mock_flumine, 123, 0.01, 100, {"test": "me"}, {"please": "now"}
        )

    def test_init(self):
        self.assertEqual(self.stream.flumine, self.mock_flumine)
        self.assertEqual(self.stream.stream_id, 123)
        self.assertEqual(self.stream.market_filter, {"test": "me"})
        self.assertEqual(self.stream.market_data_filter, {"please": "now"})
        self.assertEqual(self.stream.streaming_timeout, 0.01)
        self.assertEqual(self.stream.conflate_ms, 100)
        self.assertIsNone(self.stream._stream)

    # def test_run(self):
    #     pass
    #
    # def test_handle_output(self):
    #     pass


class TestDataStream(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_flumine = mock.Mock()
        self.stream = streams.DataStream(
            self.mock_flumine, 123, 0.01, 100, {"test": "me"}, {"please": "now"}
        )

    def test_init(self):
        self.assertEqual(self.stream.flumine, self.mock_flumine)
        self.assertEqual(self.stream.stream_id, 123)
        self.assertEqual(self.stream.market_filter, {"test": "me"})
        self.assertEqual(self.stream.market_data_filter, {"please": "now"})
        self.assertEqual(self.stream.streaming_timeout, 0.01)
        self.assertEqual(self.stream.conflate_ms, 100)
        self.assertIsNone(self.stream._stream)
        self.assertEqual(self.stream.LISTENER, datastream.FlumineListener)
        self.assertEqual(
            self.stream._listener.output_queue, self.mock_flumine.handler_queue
        )

    @mock.patch("flumine.streams.marketstream.BaseStream.betting_client")
    def test_run(self, mock_betting_client):
        self.stream.run()
        mock_betting_client.streaming.create_stream.assert_called_with(
            listener=self.stream._listener, unique_id=123
        )

    # def test_handle_output(self):
    #     pass

    @mock.patch("flumine.streams.datastream.FlumineCricketStream")
    @mock.patch("flumine.streams.datastream.FlumineRaceStream")
    @mock.patch("flumine.streams.datastream.FlumineOrderStream")
    @mock.patch("flumine.streams.datastream.FlumineMarketStream")
    def test_flumine_listener(
        self,
        mock_market_stream,
        mock_order_stream,
        mock_race_stream,
        mock_cricket_stream,
    ):
        listener = datastream.FlumineListener()
        self.assertEqual(
            listener._add_stream(123, "marketSubscription"), mock_market_stream()
        )
        self.assertEqual(
            listener._add_stream(123, "orderSubscription"), mock_order_stream()
        )
        self.assertEqual(
            listener._add_stream(123, "raceSubscription"), mock_race_stream()
        )
        self.assertEqual(
            listener._add_stream(123, "cricketSubscription"), mock_cricket_stream()
        )

    def test_flumine_stream(self):
        mock_listener = mock.Mock()
        stream = datastream.FlumineStream(mock_listener, 0)
        self.assertEqual(str(stream), "FlumineStream")
        self.assertEqual(repr(stream), "<FlumineStream [0]>")

    @mock.patch("flumine.streams.datastream.RawDataEvent")
    def test_flumine_stream_on_process(self, raw_data_event):
        mock_listener = mock.Mock()
        stream = datastream.FlumineStream(mock_listener, 0)
        stream.on_process([1, 2, 3])
        raw_data_event.assert_called_with([1, 2, 3])
        mock_listener.output_queue.put.assert_called_with(raw_data_event.return_value)

    @mock.patch("flumine.streams.datastream.FlumineMarketStream.on_process")
    def test_flumine_market_stream(self, mock_on_process):
        mock_listener = mock.Mock(stream_unique_id=0)
        stream = datastream.FlumineMarketStream(mock_listener, 0)
        stream._clk = "AAA"
        market_books = [{"id": "1.123"}, {"id": "1.456"}, {"id": "1.123"}]
        stream._process(market_books, 123)

        self.assertEqual(len(stream._caches), 2)
        self.assertEqual(stream._updates_processed, 3)
        mock_on_process.assert_called_with(
            [mock_listener.stream_unique_id, "AAA", 123, market_books]
        )

    @mock.patch("flumine.streams.datastream.FlumineMarketStream.on_process")
    def test_flumine_market_stream_market_closed(self, mock_on_process):
        mock_listener = mock.Mock(stream_unique_id=0)
        stream = datastream.FlumineMarketStream(mock_listener, 0)
        stream._caches = {"1.123": object}
        stream._clk = "AAA"
        market_books = [{"id": "1.123", "marketDefinition": {"status": "CLOSED"}}]
        stream._process(market_books, 123)

        self.assertEqual(stream._lookup, "mc")
        self.assertEqual(len(stream._caches), 0)
        self.assertEqual(stream._updates_processed, 1)
        mock_on_process.assert_called_with(
            [mock_listener.stream_unique_id, "AAA", 123, market_books]
        )

    @mock.patch("flumine.streams.datastream.FlumineOrderStream.on_process")
    def test_flumine_order_stream(self, mock_on_process):
        mock_listener = mock.Mock(stream_unique_id=0)
        stream = datastream.FlumineOrderStream(mock_listener, 0)
        stream._clk = "AAA"
        order_updates = [{"id": "1.123"}, {"id": "1.456"}, {"id": "1.123"}]
        stream._process(order_updates, 123)

        self.assertEqual(stream._lookup, "oc")
        self.assertEqual(len(stream._caches), 2)
        self.assertEqual(stream._updates_processed, 3)
        mock_on_process.assert_called_with(
            [mock_listener.stream_unique_id, "AAA", 123, order_updates]
        )

    @mock.patch("flumine.streams.datastream.FlumineRaceStream.on_process")
    def test_flumine_race_stream(self, mock_on_process):
        mock_listener = mock.Mock(stream_unique_id=0)
        stream = datastream.FlumineRaceStream(mock_listener, 0)
        stream._clk = "AAA"
        race_updates = [{"mid": "1.123"}, {"mid": "1.456"}, {"mid": "1.123"}]
        stream._process(race_updates, 123)

        self.assertEqual(stream._lookup, "rc")
        self.assertEqual(len(stream._caches), 2)
        self.assertEqual(stream._updates_processed, 3)
        mock_on_process.assert_called_with(
            [mock_listener.stream_unique_id, "AAA", 123, race_updates]
        )

    @mock.patch("flumine.streams.datastream.FlumineCricketStream.on_process")
    def test_flumine_cricket_stream(self, mock_on_process):
        mock_listener = mock.Mock(stream_unique_id=0)
        stream = datastream.FlumineCricketStream(mock_listener, 0)
        stream._clk = "AAA"
        race_updates = [
            {"marketId": "1.123"},
            {"marketId": "1.456"},
            {"marketId": "1.123"},
        ]
        stream._process(race_updates, 123)

        self.assertEqual(stream._lookup, "cc")
        self.assertEqual(len(stream._caches), 2)
        self.assertEqual(stream._updates_processed, 3)
        mock_on_process.assert_called_with(
            [mock_listener.stream_unique_id, "AAA", 123, race_updates]
        )


class TestHistoricalStream(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_flumine = mock.Mock()
        self.stream = streams.HistoricalStream(
            self.mock_flumine,
            123,
            0.01,
            100,
            {"test": "me"},
            {"please": "now"},
            inplay=True,
            seconds_to_start=123,
        )

    def test_init(self):
        self.assertEqual(self.stream.flumine, self.mock_flumine)
        self.assertEqual(self.stream.stream_id, 123)
        self.assertEqual(self.stream.market_filter, {"test": "me"})
        self.assertEqual(self.stream.market_data_filter, {"please": "now"})
        self.assertEqual(self.stream.streaming_timeout, 0.01)
        self.assertEqual(self.stream.conflate_ms, 100)
        self.assertIsNone(self.stream._stream)
        self.assertIsNone(self.stream.MAX_LATENCY)
        self.assertTrue(self.stream._listener.inplay)
        self.assertEqual(self.stream._listener.seconds_to_start, 123)

    def test_run(self):
        self.stream.run()

    def test_handle_output(self):
        self.stream.handle_output()

    @mock.patch("flumine.streams.historicalstream.FlumineHistoricalGeneratorStream")
    def test_create_generator(self, mock_generator):
        generator = self.stream.create_generator()
        mock_generator.assert_called_with(
            file_path={"test": "me"},
            listener=self.stream._listener,
            operation="marketSubscription",
            unique_id=self.stream.stream_id,
        )
        self.assertIsNone(self.stream._listener.max_latency)
        self.assertFalse(self.stream._listener.lightweight)
        self.assertFalse(self.stream._listener.update_clk)
        self.assertEqual(generator, mock_generator().get_generator())


class TestFlumineMarketStream(unittest.TestCase):
    def setUp(self) -> None:
        self.listener = mock.Mock()
        self.stream = historicalstream.FlumineMarketStream(self.listener, 0)

    def test_init(self):
        self.assertEqual(self.stream._listener, self.listener)
        self.assertEqual(self.stream._lookup, "mc")

    @mock.patch("flumine.streams.historicalstream.MarketBookCache")
    def test__process(self, mock_cache):
        update = [{"id": "1.23", "img": {1: 2}, "marketDefinition": {"runners": []}}]
        self.assertTrue(
            self.stream._process(
                update,
                12345,
            )
        )
        self.assertEqual(len(self.stream._caches), 1)
        self.assertEqual(self.stream._updates_processed, 1)
        mock_cache.assert_called_with(
            "1.23",
            12345,
            self.stream._lightweight,
            self.stream._calculate_market_tv,
            self.stream._cumulative_runner_tv,
        )
        mock_cache().update_cache.assert_called_with(update[0], 12345, active=True)

    @mock.patch("flumine.streams.historicalstream.MarketBookCache")
    def test__process_inplay(self, mock_cache):
        self.stream._listener.inplay = True
        update = [
            {
                "id": "1.23",
                "img": {1: 2},
                "marketDefinition": {"status": "OPEN", "inPlay": False, "runners": []},
            }
        ]
        self.assertFalse(
            self.stream._process(
                update,
                12345,
            )
        )
        mock_cache().update_cache.assert_called_with(update[0], 12345, active=False)

        self.stream._listener.inplay = True
        update = [
            {
                "id": "1.23",
                "img": {1: 2},
                "marketDefinition": {
                    "status": "SUSPENDED",
                    "inPlay": False,
                    "runners": [],
                },
            }
        ]
        self.assertTrue(
            self.stream._process(
                update,
                12345,
            )
        )
        mock_cache().update_cache.assert_called_with(update[0], 12345, active=True)

        update = [
            {
                "id": "1.23",
                "img": {1: 2},
                "marketDefinition": {"status": "OPEN", "inPlay": True, "runners": []},
            }
        ]
        self.assertTrue(
            self.stream._process(
                update,
                12345,
            )
        )
        mock_cache().update_cache.assert_called_with(update[0], 12345, active=True)

    @mock.patch("flumine.streams.historicalstream.MarketBookCache")
    def test__process_seconds_to_start(self, mock_cache):
        self.stream._listener.inplay = None
        self.stream._listener.seconds_to_start = 600
        update = [
            {
                "id": "1.23",
                "img": {1: 2},
                "marketDefinition": {
                    "status": "OPEN",
                    "marketTime": 456,
                    "runners": [],
                },
            }
        ]
        self.assertTrue(
            self.stream._process(
                update,
                12345,
            )
        )
        mock_cache().update_cache.assert_called_with(update[0], 12345, active=True)

        update = [
            {
                "id": "1.23",
                "img": {1: 2},
                "marketDefinition": {
                    "status": "OPEN",
                    "marketTime": 1234567,
                    "runners": [],
                },
            }
        ]
        self.assertFalse(
            self.stream._process(
                update,
                12345,
            )
        )
        mock_cache().update_cache.assert_called_with(update[0], 12345, active=False)

        update = [
            {
                "id": "1.23",
                "img": {1: 2},
                "marketDefinition": {
                    "status": "SUSPENDED",
                    "marketTime": 1234567,
                    "runners": [],
                },
            }
        ]
        self.assertTrue(
            self.stream._process(
                update,
                12345,
            )
        )
        mock_cache().update_cache.assert_called_with(update[0], 12345, active=True)

    @mock.patch("flumine.streams.historicalstream.MarketBookCache")
    def test__process_max_inplay_seconds(self, mock_cache):
        # Set listener attributes
        self.stream._listener.inplay = None
        self.stream._listener.seconds_to_start = None
        self.stream._listener.max_inplay_seconds = 10  # seconds

        # Get the mock instance
        mock_cache_instance = mock_cache.return_value
        # Initialize _definition_in_play to False
        mock_cache_instance._definition_in_play = False

        # Define side effect for update_cache
        def update_cache_side_effect(market_book, publish_time, active=True):
            if "marketDefinition" in market_book:
                market_definition = market_book["marketDefinition"]
                mock_cache_instance._definition_in_play = market_definition.get(
                    "inPlay", False
                )

        # Assign the side effect
        mock_cache_instance.update_cache.side_effect = update_cache_side_effect

        # Initial update: market not in-play
        update_pre_inplay = [
            {
                "id": "1.23",
                "img": {1: 2},
                "marketDefinition": {"status": "OPEN", "inPlay": False, "runners": []},
            }
        ]
        active = self.stream._process(update_pre_inplay, 1000)
        self.assertTrue(active)
        self.assertNotIn("1.23", self.stream.inplay_publish_times)
        mock_cache_instance.update_cache.assert_called_with(
            update_pre_inplay[0], 1000, active=True
        )

        # Market goes in-play at publish_time=2000 ms
        update_inplay_start = [
            {
                "id": "1.23",
                "img": {1: 2},
                "marketDefinition": {"status": "OPEN", "inPlay": True, "runners": []},
            }
        ]
        active = self.stream._process(update_inplay_start, 2000)
        self.assertTrue(active)
        self.assertIn("1.23", self.stream.inplay_publish_times)
        self.assertEqual(self.stream.inplay_publish_times["1.23"], 2000)
        mock_cache_instance.update_cache.assert_called_with(
            update_inplay_start[0], 2000, active=True
        )

        # Update within max_inplay_seconds (9 seconds after in-play)
        update_inplay_within_limit = [
            {
                "id": "1.23",
                "marketDefinition": {"status": "OPEN", "inPlay": True, "runners": []},
            }
        ]
        active = self.stream._process(update_inplay_within_limit, 11000)
        self.assertTrue(active)
        mock_cache_instance.update_cache.assert_called_with(
            update_inplay_within_limit[0], 11000, active=True
        )

        # Update after max_inplay_seconds (10.001 seconds after in-play)
        update_inplay_exceeded_limit = [
            {
                "id": "1.23",
                "marketDefinition": {"status": "OPEN", "inPlay": True, "runners": []},
            }
        ]
        active = self.stream._process(update_inplay_exceeded_limit, 12001)
        self.assertFalse(active)
        mock_cache_instance.update_cache.assert_called_with(
            update_inplay_exceeded_limit[0], 12001, active=False
        )


class TestFlumineRaceStream(unittest.TestCase):
    def setUp(self) -> None:
        self.listener = mock.Mock()
        self.stream = historicalstream.FlumineRaceStream(self.listener, 0)

    @mock.patch(
        "flumine.streams.historicalstream.create_time",
        return_value=datetime.datetime(1970, 1, 1, 13, 10),
    )
    def test__process(self, mock_create_time):
        self.assertTrue(
            self.stream._process(
                [{"mid": "1.23", "id": "13.10", "img": {1: 2}}],
                11111111111111,
            )
        )
        self.assertEqual(len(self.stream._caches), 1)
        self.assertEqual(self.stream._updates_processed, 1)
        self.assertTrue(self.stream._caches["1.23"].inplay)
        mock_create_time.assert_called_with(11111111111111, "13.10")

    @mock.patch(
        "flumine.streams.historicalstream.create_time",
        return_value=datetime.datetime(1970, 1, 1, 13, 10),
    )
    def test__process_false(self, mock_create_time):
        self.assertFalse(
            self.stream._process(
                [{"mid": "1.23", "id": "13.10", "img": {1: 2}}],
                1234,
            )
        )
        self.assertEqual(len(self.stream._caches), 1)
        self.assertEqual(self.stream._updates_processed, 0)
        self.assertFalse(self.stream._caches["1.23"].inplay)
        mock_create_time.assert_called_with(1234, "13.10")

    @mock.patch(
        "flumine.streams.historicalstream.create_time",
        return_value=datetime.datetime(1970, 1, 1, 13, 10),
    )
    def test__process_early_start(self, mock_create_time):
        self.assertTrue(
            self.stream._process(
                [{"mid": "1.23", "id": "13.10", "rpc": {"rt": 2.64}}],
                1,
            )
        )
        self.assertEqual(len(self.stream._caches), 1)
        self.assertEqual(self.stream._updates_processed, 1)
        self.assertTrue(self.stream._caches["1.23"].inplay)
        mock_create_time.assert_called_with(1, "13.10")


class TestFlumineCricketStream(unittest.TestCase):
    def setUp(self) -> None:
        self.listener = mock.Mock()
        self.stream = historicalstream.FlumineCricketStream(self.listener, 0)

    def test__process(self):
        self.assertTrue(
            self.stream._process(
                [{"marketId": "1.23", "eventId": "13.10"}],
                11111111111111,
            )
        )
        self.assertEqual(len(self.stream._caches), 1)
        self.assertEqual(self.stream._updates_processed, 1)


class TestHistoricListener(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_flumine = mock.Mock()
        self.listener = historicalstream.HistoricListener(
            inplay=True, seconds_to_start=123
        )

    def test_init(self):
        self.assertTrue(self.listener.inplay)
        self.assertEqual(self.listener.seconds_to_start, 123)

    @mock.patch("flumine.streams.historicalstream.FlumineMarketStream")
    def test__add_stream_market(self, mock_stream):
        self.assertEqual(
            self.listener._add_stream(123, "marketSubscription"), mock_stream()
        )

    def test__add_stream_order(self):
        with self.assertRaises(ListenerError):
            self.listener._add_stream(123, "orderSubscription")

    @mock.patch("flumine.streams.historicalstream.FlumineRaceStream")
    def test__add_stream_race(self, mock_stream):
        self.assertEqual(
            self.listener._add_stream(123, "raceSubscription"), mock_stream()
        )

    @mock.patch("flumine.streams.historicalstream.FlumineCricketStream")
    def test__add_stream_cricket(self, mock_stream):
        self.assertEqual(
            self.listener._add_stream(123, "cricketSubscription"), mock_stream()
        )

    def test__add_stream_err(self):
        with self.assertRaises(ListenerError):
            self.listener._add_stream(123, "courtSubscription")

    def test_on_data(self):
        mock_stream = mock.Mock(_lookup="mc")
        self.listener.stream = mock_stream
        self.listener.on_data('{"pt": 123, "mc": {}}')
        self.listener.stream._process.assert_called_with({}, 123)
        # error
        self.assertIsNone(self.listener.on_data("p"))


class TestOrderStream(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_flumine = mock.Mock()
        self.stream = streams.OrderStream(self.mock_flumine, 123, 0.01, 100)

    def test_init(self):
        self.assertEqual(self.stream.flumine, self.mock_flumine)
        self.assertEqual(self.stream.stream_id, 123)
        self.assertIsNone(self.stream.market_filter)
        self.assertIsNone(self.stream.market_data_filter)
        self.assertEqual(self.stream.streaming_timeout, 0.01)
        self.assertEqual(self.stream.conflate_ms, 100)
        self.assertIsNone(self.stream._stream)
        self.assertEqual(orderstream.START_DELAY, 2)
        self.assertEqual(orderstream.SNAP_DELTA, 5)

    # def test_run(self):
    #     pass
    #
    # def test_handle_output(self):
    #     pass


class TestSimulatedOrderStream(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_flumine = mock.Mock()
        self.stream = streams.SimulatedOrderStream(self.mock_flumine, 123, 0.01, 100)

    def test_init(self):
        self.assertEqual(self.stream.flumine, self.mock_flumine)
        self.assertEqual(self.stream.stream_id, 123)
        self.assertIsNone(self.stream.market_filter)
        self.assertIsNone(self.stream.market_data_filter)
        self.assertEqual(self.stream.streaming_timeout, 0.01)
        self.assertEqual(self.stream.conflate_ms, 100)
        self.assertIsNone(self.stream._stream)

    def test_current_orders(self):
        mock_client = mock.Mock()
        current_orders = CurrentOrders([1], mock_client)
        self.assertEqual(current_orders.orders, [1])
        self.assertFalse(current_orders.more_available)
        self.assertEqual(current_orders.client, mock_client)

    # def test_run(self):
    #     pass

    def test__get_current_orders(self):
        mock_market = mock.Mock(closed=False)
        order_one = mock.Mock(client=self.stream.client)
        mock_market.blotter.client_orders.return_value = [order_one]
        self.stream.flumine.markets = [mock_market, mock.Mock(closed=True)]
        self.assertEqual(self.stream._get_current_orders(), [order_one])


class TestSportsDataStream(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_flumine = mock.Mock()
        self.stream = streams.SportsDataStream(
            self.mock_flumine, 123, sports_data_filter="test", streaming_timeout=10
        )

    def test_init(self):
        self.assertEqual(self.stream.flumine, self.mock_flumine)
        self.assertEqual(self.stream.stream_id, 123)
        self.assertEqual(self.stream.sports_data_filter, "test")
        self.assertEqual(self.stream.streaming_timeout, 10)
        self.assertIsNone(self.stream._stream)
        self.assertEqual(orderstream.START_DELAY, 2)
        self.assertEqual(orderstream.SNAP_DELTA, 5)

    # def test_run(self):
    #     pass
    #
    # def test_handle_output(self):
    #     pass



import logging

from .baseflumine import BaseFlumine
from .events.events import EventType
from . import worker

logger = logging.getLogger(__name__)

MARKET_BOOK_EVENT = EventType.MARKET_BOOK
CURRENT_ORDERS_EVENT = EventType.CURRENT_ORDERS
SPORTS_DATA_EVENT = EventType.SPORTS_DATA
CUSTOM_EVENT_EVENT = EventType.CUSTOM_EVENT
RAW_DATA_EVENT = EventType.RAW_DATA
MARKET_CATALOGUE_EVENT = EventType.MARKET_CATALOGUE
CLEARED_MARKETS_EVENT = EventType.CLEARED_MARKETS
CLEARED_ORDERS_EVENT = EventType.CLEARED_ORDERS
CLOSE_MARKET_EVENT = EventType.CLOSE_MARKET
TERMINATOR_EVENT = EventType.TERMINATOR


class Flumine(BaseFlumine):
    def run(self) -> None:
        """
        Main run thread
        """
        handler_queue_get = self.handler_queue.get
        with self:
            while True:
                event = handler_queue_get()
                event_type = event.EVENT_TYPE

                if event_type == MARKET_BOOK_EVENT:
                    self._process_market_books(event)

                elif event_type == CURRENT_ORDERS_EVENT:
                    self._process_current_orders(event)

                elif event_type == CUSTOM_EVENT_EVENT:
                    self._process_custom_event(event)

                elif event_type == SPORTS_DATA_EVENT:
                    self._process_sports_data(event)

                elif event_type == RAW_DATA_EVENT:
                    self._process_raw_data(event)

                elif event_type == MARKET_CATALOGUE_EVENT:
                    self._process_market_catalogues(event)

                elif event_type == CLEARED_MARKETS_EVENT:
                    self._process_cleared_markets(event)

                elif event_type == CLEARED_ORDERS_EVENT:
                    self._process_cleared_orders(event)

                elif event_type == CLOSE_MARKET_EVENT:
                    self._process_close_market(event)

                elif event_type == TERMINATOR_EVENT:
                    break

                else:
                    logger.error("Unknown item in handler_queue: %s" % str(event))

                del event

    def _add_default_workers(self):
        client_timeouts = [
            client.betting_client.session_timeout for client in self.clients
        ]
        ka_interval = min((min(client_timeouts) / 2), 1200)
        self.add_worker(
            worker.BackgroundWorker(
                self, function=worker.keep_alive, interval=ka_interval
            )
        )
        self.add_worker(
            worker.BackgroundWorker(
                self,
                function=worker.poll_market_catalogue,
                interval=60,
                start_delay=10,  # wait for streams to populate
            )
        )
        if not all([client.market_recording_mode for client in self.clients]):
            self.add_worker(
                worker.BackgroundWorker(
                    self,
                    function=worker.poll_account_balance,
                    interval=120,
                    start_delay=10,  # wait for login
                )
            )
            self.add_worker(
                worker.BackgroundWorker(
                    self,
                    function=worker.poll_market_closure,
                    interval=60,
                    start_delay=10,  # wait for login
                )
            )

    def __repr__(self) -> str:
        return "<Flumine>"

    def __str__(self) -> str:
        return "<Flumine>"



import unittest
import datetime
from unittest import mock
from collections import defaultdict

from flumine.markets.markets import Markets
from flumine.markets.market import Market
from flumine import config


class MarketsTest(unittest.TestCase):
    def setUp(self) -> None:
        self.markets = Markets()

    def test_init(self):
        self.assertEqual(self.markets._markets, {})
        self.assertEqual(self.markets.events, {})

    def test_add_market(self):
        mock_market = mock.Mock(event_id="1234")
        self.markets.add_market("1.1", mock_market)
        self.assertEqual(self.markets._markets, {"1.1": mock_market})
        self.assertEqual(self.markets.events, {"1234": [mock_market]})

    def test_add_market_no_event_id(self):
        mock_market = mock.Mock(event_id=None)
        self.markets.add_market("1.1", mock_market)
        self.assertEqual(self.markets._markets, {"1.1": mock_market})
        self.assertEqual(self.markets.events, {})

    def test_add_market_reopen(self):
        mock_market = mock.Mock()
        self.markets._markets = {"1.1": mock_market}
        self.markets.add_market("1.1", mock_market)

        self.assertEqual(self.markets._markets, {"1.1": mock_market})
        mock_market.open_market.assert_called_with()

    def test_close_market(self):
        mock_market = mock.Mock()
        self.markets._markets = {"1.1": mock_market}
        self.markets.close_market("1.1")
        mock_market.close_market.assert_called_with()

    def test_remove_market(self):
        mock_market = mock.Mock(event_id=1234)
        self.markets._markets = {"1.1": mock_market}
        self.markets.events = {1234: [mock_market]}
        self.markets.remove_market("1.1")
        self.assertEqual(self.markets._markets, {})
        self.assertEqual(self.markets.events, {1234: []})

    def test_remove_market_no_event(self):
        mock_market = mock.Mock(event_id=1234)
        self.markets._markets = {"1.1": mock_market}
        self.markets.remove_market("1.1")
        self.assertEqual(self.markets._markets, {})
        self.assertEqual(self.markets.events, {})

    def test_remove_market_err(self):
        mock_market = mock.Mock(event_id=1234)
        mock_market_two = mock.Mock(event_id=1234)
        self.markets._markets = {"1.1": mock_market, "2.2": mock_market_two}
        self.markets.events = {1234: [mock_market]}
        self.markets.remove_market("2.2")
        self.assertEqual(self.markets._markets, {"1.1": mock_market})
        self.assertEqual(self.markets.events, {1234: [mock_market]})

    def test_get_order(self):
        mock_market = mock.Mock()
        mock_market.closed = False
        mock_market.blotter = {"test": 12}
        self.markets._markets = {"1.1": mock_market}

        self.assertEqual(self.markets.get_order("1.1", "test"), 12)
        self.assertIsNone(self.markets.get_order("1.2", "test"))

    def test_get_order_from_bet_id(self):
        mock_blotter = mock.Mock()
        mock_market = mock.Mock(closed=False, blotter=mock_blotter)
        self.markets._markets = {"1.1": mock_market}
        self.markets.get_order_from_bet_id("1.1", "321")
        mock_blotter.get_order_bet_id.assert_called_with("321")

    def test_markets(self):
        self.assertEqual(self.markets.markets, {})
        mock_market = mock.Mock()
        mock_market_two = mock.Mock()
        self.markets._markets = {"1.1": mock_market, "2.1": mock_market_two}
        self.assertEqual(
            self.markets.markets, {"1.1": mock_market, "2.1": mock_market_two}
        )

    def test_open_market_ids(self):
        self.assertEqual(self.markets.open_market_ids, [])
        mock_market = mock.Mock(status="OPEN")
        mock_market_two = mock.Mock(status="CLOSED")
        self.markets._markets = {"1.1": mock_market, "2.1": mock_market_two}
        self.assertEqual(self.markets.open_market_ids, [mock_market.market_id])

    def test_live_orders(self):
        self.assertFalse(self.markets.live_orders)
        mock_market = mock.Mock()
        mock_market.closed = False
        mock_market.blotter.has_live_orders = True
        self.markets._markets = {"1.234": mock_market}
        self.assertTrue(self.markets.live_orders)
        mock_market.blotter.has_live_orders = False
        self.assertFalse(self.markets.live_orders)

    def test_iter(self):
        self.assertEqual(len([i for i in self.markets]), 0)

    def test_len(self):
        self.assertEqual(len(self.markets), 0)


class MarketTest(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_flumine = mock.Mock()
        self.mock_market_book = mock.Mock()
        self.mock_market_catalogue = mock.Mock()
        self.market = Market(
            self.mock_flumine,
            "1.234",
            self.mock_market_book,
            self.mock_market_catalogue,
        )

    def test_init(self):
        self.assertEqual(self.market.flumine, self.mock_flumine)
        self.assertEqual(self.market.market_id, "1.234")
        self.assertFalse(self.market.closed)
        self.assertIsNotNone(self.market.date_time_created)
        self.assertIsNone(self.market.date_time_closed)
        self.assertEqual(self.market.market_book, self.mock_market_book)
        self.assertEqual(self.market.market_catalogue, self.mock_market_catalogue)
        self.assertTrue(self.market.update_market_catalogue)
        self.assertEqual(self.market.orders_cleared, [])
        self.assertEqual(self.market.market_cleared, [])
        self.assertEqual(self.market.context, {"simulated": {}})
        self.assertIsNotNone(self.market.blotter)
        self.assertEqual(self.market._transaction_id, 0)

    def test_call(self):
        mock_market_book = mock.Mock()
        self.market(mock_market_book)
        self.assertEqual(self.market.market_book, mock_market_book)
        self.assertTrue(self.market.update_market_catalogue)

    def test_open_market(self):
        self.market.closed = True
        self.market.orders_cleared = [1, 2]
        self.market.market_cleared = [1]
        self.market.open_market()
        self.assertFalse(self.market.closed)
        self.assertEqual(self.market.orders_cleared, [])
        self.assertEqual(self.market.market_cleared, [])

    def test_close_market(self):
        self.market.close_market()
        self.assertTrue(self.market.closed)
        self.assertIsNotNone(self.market.date_time_closed)

    @mock.patch("flumine.markets.market.Transaction")
    def test_transaction(self, mock_transaction):
        transaction = self.market.transaction()
        mock_transaction.assert_called_with(
            self.market,
            id_=self.market._transaction_id,
            async_place_orders=False,
            client=self.market.flumine.clients.get_default(),
        )
        self.assertEqual(transaction, mock_transaction())

    @mock.patch("flumine.markets.market.Transaction")
    def test_transaction_async(self, mock_transaction):
        config.async_place_orders = True
        transaction = self.market.transaction()
        mock_transaction.assert_called_with(
            self.market,
            id_=self.market._transaction_id,
            async_place_orders=True,
            client=self.market.flumine.clients.get_default(),
        )
        self.assertEqual(transaction, mock_transaction())

    @mock.patch("flumine.markets.market.Market.transaction")
    def test_place_order(self, mock_transaction):
        mock_transaction.return_value.__enter__.return_value = mock_transaction
        mock_order = mock.Mock()
        self.assertTrue(
            self.market.place_order(mock_order, 2, False, force=True, client=1)
        )
        mock_transaction.assert_called_with(client=1)
        mock_transaction.place_order.assert_called_with(mock_order, 2, False, True)

    @mock.patch("flumine.markets.market.Market.transaction")
    def test_cancel_order(self, mock_transaction):
        mock_transaction.return_value.__enter__.return_value = mock_transaction
        mock_order = mock.Mock()
        self.assertTrue(self.market.cancel_order(mock_order, 2.02, force=True))
        mock_transaction.assert_called_with(client=mock_order.client)
        mock_transaction.cancel_order.assert_called_with(mock_order, 2.02, True)

    @mock.patch("flumine.markets.market.Market.transaction")
    def test_update_order(self, mock_transaction):
        mock_transaction.return_value.__enter__.return_value = mock_transaction
        mock_order = mock.Mock()
        self.assertTrue(self.market.update_order(mock_order, "test", force=True))
        mock_transaction.assert_called_with(client=mock_order.client)
        mock_transaction.update_order.assert_called_with(mock_order, "test", True)

    @mock.patch("flumine.markets.market.Market.transaction")
    def test_replace_order(self, mock_transaction):
        mock_transaction.return_value.__enter__.return_value = mock_transaction
        mock_order = mock.Mock()
        self.assertTrue(self.market.replace_order(mock_order, 2, False, force=True))
        mock_transaction.assert_called_with(client=mock_order.client)
        mock_transaction.replace_order.assert_called_with(mock_order, 2, False, True)

    def test_event(self):
        self.market.market_catalogue.event.id = 12
        self.market.market_catalogue.event_type.id = "7"
        self.market.market_book.market_definition.market_time = 12
        self.market.flumine.markets.events = defaultdict(list)
        self.assertEqual(self.market.event, {})

        m_one = mock.Mock(market_type=1, event_id=12, market_start_datetime=12)
        m_two = mock.Mock(market_type=2, event_id=12, market_start_datetime=12)
        m_three = mock.Mock(market_type=3, event_id=123, market_start_datetime=12)
        m_four = mock.Mock(market_type=1, event_id=12, market_start_datetime=12)
        m_five = mock.Mock(market_type=2, event_id=12, market_start_datetime=13)
        m_six = mock.Mock(
            market_type=2, event_id=13, market_start_datetime=13
        )  # different event
        self.market.flumine.markets.events[m_one.event_id].append(m_one)
        self.market.flumine.markets.events[m_two.event_id].append(m_two)
        self.market.flumine.markets.events[m_three.event_id].append(m_three)
        self.market.flumine.markets.events[m_four.event_id].append(m_four)
        self.market.flumine.markets.events[m_five.event_id].append(m_five)
        self.market.flumine.markets.events[m_six.event_id].append(m_six)
        self.assertEqual(self.market.event, {1: [m_one, m_four], 2: [m_two]})

    def test_event_filters(self):
        self.market.market_catalogue.event.id = 12
        self.market.market_catalogue.event_type.id = "1"
        self.market.market_book.market_definition.market_time = 12
        self.market.flumine.markets.events = defaultdict(list)
        self.assertEqual(self.market.event, {})

        m_one = mock.Mock(
            market_type=1, event_id=12, event_type_id="1", market_start_datetime=12
        )
        m_two = mock.Mock(
            market_type=2, event_id=12, event_type_id="1", market_start_datetime=12
        )
        m_three = mock.Mock(
            market_type=3, event_id=123, event_type_id="1", market_start_datetime=12
        )
        m_four = mock.Mock(
            market_type=2, event_id=12, event_type_id="1", market_start_datetime=13
        )
        m_five = mock.Mock(
            market_type=1, event_id=13, event_type_id="1", market_start_datetime=13
        )
        m_six = mock.Mock(
            market_type=2, event_id=13, event_type_id="3", market_start_datetime=13
        )
        self.market.flumine.markets.events[m_one.event_id].append(m_one)
        self.market.flumine.markets.events[m_two.event_id].append(m_two)
        self.market.flumine.markets.events[m_three.event_id].append(m_three)
        self.market.flumine.markets.events[m_four.event_id].append(m_four)
        self.market.flumine.markets.events[m_five.event_id].append(m_five)
        self.market.flumine.markets.events[m_six.event_id].append(m_six)
        self.assertEqual(self.market.event, {1: [m_one], 2: [m_two, m_four]})

        self.market.market_catalogue.event.id = 13
        self.market.market_catalogue.event_type.id = "3"
        self.market.market_book.market_definition.market_time = 12
        self.market.flumine.markets.events = defaultdict(list)
        m_six = mock.Mock(
            market_type=2, event_id=13, event_type_id="3", market_start_datetime=13
        )
        self.market.flumine.markets.events[m_six.event_id].append(m_six)
        self.assertEqual(self.market.event, {2: [m_six]})

    def test_event_type_id_mc(self):
        mock_market_catalogue = mock.Mock()
        self.market.market_catalogue = mock_market_catalogue
        self.assertEqual(self.market.event_type_id, mock_market_catalogue.event_type.id)

    def test_event_type_id_mb(self):
        self.market.market_catalogue = None
        mock_market_book = mock.Mock()
        self.market.market_book = mock_market_book
        self.assertEqual(
            self.market.event_type_id, mock_market_book.market_definition.event_type_id
        )

    def test_event_id_mc(self):
        mock_market_catalogue = mock.Mock()
        self.market.market_catalogue = mock_market_catalogue
        self.assertEqual(self.market.event_id, mock_market_catalogue.event.id)

    def test_event_id_mb(self):
        self.market.market_catalogue = None
        mock_market_book = mock.Mock()
        self.market.market_book = mock_market_book
        self.assertEqual(
            self.market.event_id, mock_market_book.market_definition.event_id
        )

    def test_market_type_mc(self):
        mock_market_catalogue = mock.Mock()
        self.market.market_catalogue = mock_market_catalogue
        self.assertEqual(
            self.market.market_type, mock_market_catalogue.description.market_type
        )
        self.market.market_catalogue.description = None
        mock_market_book = mock.Mock()
        self.market.market_book = mock_market_book
        self.assertEqual(
            self.market.market_type, mock_market_book.market_definition.market_type
        )

    def test_market_type_mb(self):
        self.market.market_catalogue = None
        mock_market_book = mock.Mock()
        self.market.market_book = mock_market_book
        self.assertEqual(
            self.market.market_type, mock_market_book.market_definition.market_type
        )

    def test_seconds_to_start(self):
        self.market.market_book = None
        mock_market_catalogue = mock.Mock()
        mock_market_catalogue.market_start_time = datetime.datetime.utcfromtimestamp(1)
        self.market.market_catalogue = mock_market_catalogue
        self.assertLess(self.market.seconds_to_start, 0)

    def test_seconds_to_start_market_book(self):
        self.market.market_catalogue = None
        mock_market_book = mock.Mock()
        mock_market_book.market_definition.market_time = (
            datetime.datetime.utcfromtimestamp(1)
        )
        self.market.market_book = mock_market_book
        self.assertLess(self.market.seconds_to_start, 0)

    def test_seconds_to_start_market_catalogue(self):
        self.market.market_book.market_definition.market_time = (
            datetime.datetime.utcfromtimestamp(1)
        )
        self.assertLess(self.market.seconds_to_start, 0)

    def test_seconds_to_start_none(self):
        self.market.market_book = None
        self.market.market_catalogue = None
        self.assertLess(self.market.seconds_to_start, 0)

    def test_elapsed_seconds_closed(self):
        self.assertIsNone(self.market.elapsed_seconds_closed)
        self.market.closed = True
        self.market.date_time_closed = datetime.datetime.utcnow()
        self.assertGreaterEqual(self.market.elapsed_seconds_closed, 0)

    def test_market_start_datetime(self):
        self.assertEqual(
            self.market.market_start_datetime,
            self.mock_market_book.market_definition.market_time,
        )
        self.market.market_book = None
        self.assertEqual(
            self.market.market_start_datetime,
            self.mock_market_catalogue.market_start_time,
        )
        self.market.market_catalogue = None
        self.assertEqual(
            self.market.market_start_datetime, datetime.datetime.utcfromtimestamp(0)
        )

    @mock.patch(
        "flumine.markets.market.Market.market_start_datetime",
        new_callable=mock.PropertyMock,
        return_value=None,
    )
    def test_market_start_hour_minute(self, mock_market_start_datetime):
        self.assertIsNone(self.market.market_start_hour_minute)
        mock_market_start_datetime.return_value = datetime.datetime.utcfromtimestamp(
            12000000000
        )
        self.assertEqual(self.market.market_start_hour_minute, "2120")

    def test_event_name_mc(self):
        mock_market_catalogue = mock.Mock()
        self.market.market_catalogue = mock_market_catalogue
        self.assertEqual(self.market.event_name, mock_market_catalogue.event.name)

    def test_event_name_mb(self):
        self.market.market_catalogue = None
        mock_market_book = mock.Mock()
        self.market.market_book = mock_market_book
        self.assertEqual(
            self.market.event_name, mock_market_book.market_definition.event_name
        )

    def test_country_code_mc(self):
        mock_market_catalogue = mock.Mock()
        self.market.market_catalogue = mock_market_catalogue
        self.assertEqual(
            self.market.country_code, mock_market_catalogue.event.country_code
        )

    def test_country_code_mb(self):
        self.market.market_catalogue = None
        mock_market_book = mock.Mock()
        self.market.market_book = mock_market_book
        self.assertEqual(
            self.market.country_code, mock_market_book.market_definition.country_code
        )

    def test_venue_mc(self):
        mock_market_catalogue = mock.Mock()
        self.market.market_catalogue = mock_market_catalogue
        self.assertEqual(self.market.venue, mock_market_catalogue.event.venue)

    def test_venue_mb(self):
        self.market.market_catalogue = None
        mock_market_book = mock.Mock()
        self.market.market_book = mock_market_book
        self.assertEqual(self.market.venue, mock_market_book.market_definition.venue)

    def test_race_type_mc(self):
        mock_market_catalogue = mock.Mock()
        self.market.market_catalogue = mock_market_catalogue
        self.assertEqual(
            self.market.race_type, mock_market_catalogue.description.race_type
        )
        self.market.market_catalogue.description = None
        mock_market_book = mock.Mock()
        self.market.market_book = mock_market_book
        self.assertEqual(
            self.market.race_type, mock_market_book.market_definition.race_type
        )

    def test_race_type_mb(self):
        self.market.market_catalogue = None
        mock_market_book = mock.Mock()
        self.market.market_book = mock_market_book
        self.assertEqual(
            self.market.race_type, mock_market_book.market_definition.race_type
        )

    def test_status(self):
        mock_market_book = mock.Mock(status="OPEN")
        self.market.market_book = mock_market_book
        self.assertEqual(self.market.status, mock_market_book.status)

    @mock.patch("flumine.markets.market.Market.event_type_id")
    @mock.patch("flumine.markets.market.Market.event_id")
    def test_cleared(self, mock_event_id, mock_event_type_id):
        mock_client = mock.Mock(commission_base=0.05)
        mock_order = mock.Mock(
            client=mock_client, size_matched=2.00, profit=20, lookup=(1, 2, 3)
        )
        mock_order.simulated.profit = 20
        self.market.blotter["123"] = mock_order
        self.assertEqual(
            self.market.cleared(mock_client),
            {
                "betCount": 1,
                "betOutcome": "WON",
                "commission": 1.0,
                "customerStrategyRef": config.customer_strategy_ref,
                "eventId": mock_event_id,
                "eventTypeId": mock_event_type_id,
                "lastMatchedDate": None,
                "marketId": "1.234",
                "placedDate": None,
                "profit": 20,
                "settledDate": None,
            },
        )

    def test_info(self):
        self.assertEqual(
            self.market.info,
            {
                "market_id": self.market.market_id,
                "event_id": self.market.event_id,
                "event_type_id": self.market.event_type_id,
                "event_name": self.market.event_name,
                "market_type": self.market.market_type,
                "market_start_datetime": str(self.market.market_start_datetime),
                "country_code": self.market.country_code,
                "venue": self.market.venue,
                "race_type": self.market.race_type,
                "orders_cleared": self.market.orders_cleared,
                "market_cleared": self.market.market_cleared,
                "closed": self.market.closed,
            },
        )



import logging
import datetime
import smart_open
from typing import Optional
from betfairlightweight.streaming import StreamListener, HistoricalGeneratorStream
from betfairlightweight.streaming.stream import MarketStream, RaceStream, CricketStream
from betfairlightweight.streaming.cache import (
    MarketBookCache,
    RaceCache,
    CricketMatchCache,
)
from betfairlightweight.resources.baseresource import BaseResource
from betfairlightweight.compat import json

from .basestream import BaseStream
from ..exceptions import ListenerError
from ..utils import create_time

logger = logging.getLogger(__name__)


class FlumineMarketStream(MarketStream):
    """
    Custom bflw stream to speed up processing
    by limiting to inplay/not inplay or limited
    seconds to start.
    `_process` updated to not call `on_process`
    which reduces some function calls.
    """

    def __init__(self, *args, **kwargs):
        super(FlumineMarketStream, self).__init__(*args, **kwargs)
        self.inplay_publish_times = {}

    def _process(self, data: list, publish_time: int) -> bool:
        active = False
        for market_book in data:
            if "id" not in market_book:
                continue
            market_id = market_book["id"]
            full_image = market_book.get("img", False)
            market_book_cache = self._caches.get(market_id)

            if (
                full_image or market_book_cache is None
            ):  # historic data does not contain img
                if "marketDefinition" not in market_book:
                    logger.warning(
                        "[%s: %s]: Missing marketDefinition on market %s resulting "
                        "in potential missing data in the MarketBook (make sure "
                        "EX_MARKET_DEF is requested)",
                        self,
                        self.unique_id,
                        market_id,
                    )
                market_book_cache = MarketBookCache(
                    market_id,
                    publish_time,
                    self._lightweight,
                    self._calculate_market_tv,
                    self._cumulative_runner_tv,
                )
                self._caches[market_id] = market_book_cache
                logger.info(
                    "[%s: %s]: %s added, %s markets in cache",
                    self,
                    self.unique_id,
                    market_id,
                    len(self._caches),
                )

            # listener_kwargs filtering
            active = True
            if "marketDefinition" in market_book:
                _definition_status = market_book["marketDefinition"].get("status")
                _definition_in_play = market_book["marketDefinition"].get("inPlay")
                _definition_market_time = market_book["marketDefinition"].get(
                    "marketTime"
                )
            else:
                _definition_status = market_book_cache._definition_status
                _definition_in_play = market_book_cache._definition_in_play
                _definition_market_time = market_book_cache.market_definition[
                    "marketTime"
                ]

            if (
                self._listener.max_inplay_seconds is not None
                and _definition_in_play
                and not market_book_cache._definition_in_play
            ):
                self.inplay_publish_times[market_id] = publish_time

            # if market is not open (closed/suspended) process regardless
            if _definition_status == "OPEN":
                if self._listener.inplay:
                    if not _definition_in_play:
                        active = False
                elif self._listener.seconds_to_start:
                    _now = datetime.datetime.utcfromtimestamp(publish_time / 1e3)
                    _market_time = BaseResource.strip_datetime(_definition_market_time)
                    seconds_to_start = (_market_time - _now).total_seconds()
                    if seconds_to_start > self._listener.seconds_to_start:
                        active = False
                if self._listener.inplay is False:
                    if _definition_in_play:
                        active = False
                if (
                    self._listener.max_inplay_seconds is not None
                    and market_id in self.inplay_publish_times
                ):
                    inplay_seconds = (
                        publish_time - self.inplay_publish_times[market_id]
                    ) / 1000
                    if inplay_seconds > self._listener.max_inplay_seconds:
                        active = False
            # check if refresh required
            if active and not market_book_cache.active:
                market_book_cache.refresh_cache()

            market_book_cache.update_cache(market_book, publish_time, active=active)
            self._updates_processed += 1
        return active


class FlumineRaceStream(RaceStream):
    """
    `_process` updated to not call `on_process`
    which reduces some function calls.
    """

    def _process(self, race_updates: list, publish_time: int) -> bool:
        active = False
        for update in race_updates:
            market_id = update["mid"]
            race_cache = self._caches.get(market_id)
            if race_cache is None:
                race_id = update.get("id")
                race_cache = RaceCache(
                    market_id, publish_time, race_id, self._lightweight
                )
                race_cache.start_time = create_time(publish_time, race_id)
                race_cache.inplay = False
                self._caches[market_id] = race_cache
                logger.info(
                    "[%s: %s]: %s added, %s markets in cache",
                    self,
                    self.unique_id,
                    market_id,
                    len(self._caches),
                )

            if race_cache.inplay:
                race_cache.update_cache(update, publish_time)
                self._updates_processed += 1
                active = True
            else:
                # filter after start time
                diff = (
                    race_cache.start_time
                    - datetime.datetime.utcfromtimestamp(publish_time / 1e3)
                ).total_seconds()
                # handle US races starting early by checking the update
                if diff <= 0 or (
                    "rpc" in update
                    and (
                        update["rpc"]["rt"]
                        or update["rpc"]["spd"]
                        or update["rpc"]["ord"]
                    )
                ):
                    race_cache.inplay = True
                    race_cache.update_cache(update, publish_time)
                    self._updates_processed += 1
                    active = True
        return active


class FlumineCricketStream(CricketStream):
    """
    `_process` updated to not call `on_process`
    which reduces some function calls.
    """

    def _process(self, cricket_changes: list, publish_time: int) -> bool:
        caches = []
        for cricket_change in cricket_changes:
            market_id = cricket_change["marketId"]
            event_id = cricket_change["eventId"]
            cricket_match_cache = self._caches.get(market_id)

            if cricket_match_cache is None:
                cricket_match_cache = CricketMatchCache(
                    market_id, event_id, publish_time, self._lightweight
                )
                self._caches[market_id] = cricket_match_cache
                logger.info(
                    "[%s: %s]: %s added, %s markets in cache",
                    self,
                    self.unique_id,
                    market_id,
                    len(self._caches),
                )

            cricket_match_cache.update_cache(cricket_change, publish_time)
            caches.append(cricket_match_cache)
            self._updates_processed += 1
        return True


class HistoricListener(StreamListener):
    """
    Custom listener to restrict processing by
    inplay or seconds_to_start.
    """

    def __init__(
        self,
        inplay: Optional[bool] = None,
        seconds_to_start: Optional[float] = None,
        max_inplay_seconds: Optional[float] = None,
        **kwargs,
    ):
        super(HistoricListener, self).__init__(**kwargs)
        self.inplay = inplay
        self.seconds_to_start = seconds_to_start
        self.max_inplay_seconds = max_inplay_seconds

    def _add_stream(self, unique_id: int, operation: str):
        if operation == "marketSubscription":
            return FlumineMarketStream(self, unique_id)
        elif operation == "orderSubscription":
            raise ListenerError("Unable to process order stream")
        elif operation == "raceSubscription":
            return FlumineRaceStream(self, unique_id)
        elif operation == "cricketSubscription":
            return FlumineCricketStream(self, unique_id)
        else:
            raise ListenerError("Unable to process '{0}' stream".format(operation))

    def on_data(self, raw_data: bytes) -> Optional[bool]:
        try:
            data = json.loads(raw_data)
        except ValueError:
            logger.error("value error: %s" % raw_data)
            return

        # remove error handler / operation check

        # skip on_change / on_update as we know it is always an update
        publish_time = data["pt"]
        return self.stream._process(data[self.stream._lookup], publish_time)


class FlumineHistoricalGeneratorStream(HistoricalGeneratorStream):
    """Super fast historical stream"""

    def _read_loop(self) -> dict:
        unique_id = self.unique_id
        self.listener.register_stream(unique_id, self.operation)
        listener_on_data = self.listener.on_data  # cache functions
        caches = self.listener.stream._caches
        with smart_open.open(self.file_path, "r") as f:
            file = f.readlines()  # read entire file into memory (faster)
            for update in file:
                if listener_on_data(update):
                    yield [
                        cache.create_resource(unique_id, snap=True)
                        for cache in caches.values()
                        if cache.active
                    ]


class HistoricalStream(BaseStream):
    LISTENER = HistoricListener
    MAX_LATENCY = None

    def run(self) -> None:
        pass

    def handle_output(self) -> None:
        pass

    def create_generator(self):
        self._listener.update_clk = (
            False  # do not update clk on updates (not required when simulating)
        )
        stream = FlumineHistoricalGeneratorStream(
            file_path=self.market_filter,
            listener=self._listener,
            operation=self.operation,
            unique_id=self.stream_id,
        )
        return stream.get_generator()



import time
import queue
import logging
from betfairlightweight import BetfairError, filters
from tenacity import retry

from .basestream import BaseStream
from ..events.events import CurrentOrdersEvent
from .. import config

logger = logging.getLogger(__name__)

RETRY_WAIT = BaseStream.RETRY_WAIT
START_DELAY = 2
SNAP_DELTA = 5


class OrderStream(BaseStream):
    @retry(wait=RETRY_WAIT)
    def run(self) -> None:
        logger.info(
            "Starting OrderStream %s",
            self.stream_id,
            extra={
                "stream_id": self.stream_id,
                "customer_strategy_refs": config.customer_strategy_ref,
                "conflate_ms": self.conflate_ms,
                "streaming_timeout": self.streaming_timeout,
                "client_username": self.client.username,
            },
        )
        if not self._output_thread.is_alive():
            logger.info("Starting output_thread (OrderStream %s)", self.stream_id)
            self._output_thread.start()

        self._stream = self.betting_client.streaming.create_stream(
            unique_id=self.stream_id, listener=self._listener
        )
        try:
            self.stream_id = self._stream.subscribe_to_orders(
                order_filter=filters.streaming_order_filter(
                    customer_strategy_refs=(
                        [config.customer_strategy_ref]
                        if config.customer_strategy_ref
                        else None
                    ),
                    partition_matched_by_strategy_ref=True,
                    include_overall_position=False,
                ),
                conflate_ms=self.conflate_ms,
            )
            self._stream.start()
        except BetfairError:
            logger.error("OrderStream %s run error", self.stream_id, exc_info=True)
            raise
        except Exception:
            logger.critical("OrderStream %s run error", self.stream_id, exc_info=True)
            raise
        logger.info("Stopped OrderStream %s", self.stream_id)

    def handle_output(self) -> None:
        """Handles output from stream, snaps
        if live orders or time delta greater
        than SNAP_DELTA.
        """
        last_snap = 1
        while self.is_alive():
            try:
                order_books = self._output_queue.get(
                    block=True, timeout=self.streaming_timeout
                )
            except queue.Empty:
                active_open_markets = [
                    m
                    for m in self.flumine.markets
                    if m.blotter.active and m.status == "OPEN"
                ]
                if active_open_markets or (time.time() - last_snap) > SNAP_DELTA:
                    order_books = []
                else:
                    continue
            last_snap = time.time()
            if order_books or self.flumine.markets.live_orders:
                for order_book in order_books:
                    order_book.client = self.client
                self.flumine.handler_queue.put(CurrentOrdersEvent(order_books))

        logger.info("Stopped output_thread (OrderStream %s)", self.stream_id)



import os
import socket

simulated = False
simulated_strategy_isolation = True
simulation_available_prices = False

instance_id = None  # instance id (e.g. AWS ec2 instanceId)

customer_strategy_ref = socket.gethostname()[
    :15
]  # ie. docker container id (used as order customerStrategyRefs)

hostname = customer_strategy_ref

process_id = os.getpid()  # process id of app

current_time = None  # used for simulation

raise_errors = False  # used for call_check_market / call_process_market_book

max_execution_workers = 32  # max number of workers in execution thread pool

async_place_orders = False  # async place orders

# latencies used for simulation
place_latency = 0.120
cancel_latency = 0.170
update_latency = 0.150
replace_latency = 0.280

order_sep = "-"  # customer_order_ref separator

execution_retry_attempts = 10  # cancel attempts when the OrderStream is not connected



class FlumineException(Exception):
    """Base class for Flumine Errors"""

    pass


class RunError(FlumineException):
    """Exception raised if error
    in `Flumine.run()``
    """

    def __init__(self, message):
        super(RunError, self).__init__(message)


class ListenerError(FlumineException):
    """Error raised if error in Listener"""

    def __int__(self, message):
        super(ListenerError, self).__init__(message)


class OrderError(FlumineException):
    """Exception raised if incorrect
    order/order_type requested.
    """

    def __init__(self, message):
        super(OrderError, self).__init__(message)


class OrderUpdateError(FlumineException):
    """Exception raised if order update
    incorrect.
    """

    def __init__(self, message):
        super(OrderUpdateError, self).__init__(message)


class OrderExecutionError(FlumineException):
    """Exception raised error in package during
    execution.
    """

    pass


class ControlError(FlumineException):
    """Exception raised if order voilates
    a control.
    """

    def __init__(self, message):
        super(ControlError, self).__init__(message)


class ClientError(FlumineException):
    """Exception raised on client
    error.
    """

    def __init__(self, message):
        super(ClientError, self).__init__(message)



import logging
import unittest
from unittest import mock
from betfairlightweight import BetfairError, exceptions

from flumine import worker
from flumine.clients import ExchangeType


class BackgroundWorkerTest(unittest.TestCase):
    def setUp(self):
        self.mock_function = mock.Mock(__name__="test")
        self.mock_flumine = mock.Mock()
        self.worker = worker.BackgroundWorker(
            self.mock_flumine,
            self.mock_function,
            0,
            (1, 2),
            {"hello": "world"},
            0,
            {1: 2},
        )

    def test_init(self):
        self.assertEqual(self.worker.interval, 0)
        self.assertEqual(self.worker.function, self.mock_function)
        self.assertEqual(self.worker.flumine, self.mock_flumine)
        self.assertEqual(self.worker.func_args, (1, 2))
        self.assertEqual(self.worker.func_kwargs, {"hello": "world"})
        self.assertEqual(self.worker.start_delay, 0)
        self.assertEqual(self.worker.context, {1: 2})
        self.assertEqual(self.worker.name, "test")
        self.assertFalse(self.worker._running)

    def test_run_none(self):
        worker_ = worker.BackgroundWorker(
            self.mock_flumine,
            self.mock_function,
            None,
            (1, 2),
            {"hello": "world"},
            0,
            {1: 2},
        )
        worker_.run()
        self.mock_function.assert_called_with(
            {1: 2}, self.mock_flumine, 1, 2, hello="world"
        )

    def test_shutdown(self):
        self.worker.start()
        while not self.worker._running:
            continue  # wait for thread to start
        self.assertTrue(self.worker.is_alive())
        self.worker.shutdown()
        self.assertFalse(self.worker._running)
        self.assertFalse(self.worker.is_alive())


class WorkersTest(unittest.TestCase):
    def setUp(self) -> None:
        logging.disable(logging.CRITICAL)

    def test_keep_alive(self):
        mock_context = mock.Mock()
        mock_client_simulated = mock.Mock(EXCHANGE=ExchangeType.SIMULATED)
        mock_client_betfair = mock.Mock(EXCHANGE=ExchangeType.BETFAIR)
        mock_client_betfair.betting_client.session_token = None
        mock_client_betfair_st = mock.Mock(EXCHANGE=ExchangeType.BETFAIR)
        mock_client_betfair_st.betting_client.session_token = "test"
        mock_client_betconnect = mock.Mock(EXCHANGE=ExchangeType.BETCONNECT)
        mock_client_betconnect.keep_alive.return_value = None
        mock_flumine = mock.Mock(
            clients=[
                mock_client_simulated,
                mock_client_betfair,
                mock_client_betfair_st,
                mock_client_betconnect,
            ]
        )
        worker.keep_alive(mock_context, mock_flumine)

        mock_client_simulated.login.assert_called_with()
        mock_client_betfair.login.assert_called_with()
        mock_client_betfair_st.keep_alive.assert_called_with()
        mock_client_betfair_st.login.assert_called_with()
        mock_client_betconnect.keep_alive.assert_called_with()
        mock_client_betconnect.login.assert_called_with()

    @mock.patch("flumine.worker.events")
    def test_poll_market_catalogue(self, mock_events):
        mock_context = mock.Mock()
        mock_flumine = mock.Mock()
        mock_client = mock.Mock()
        mock_flumine.clients.get_betfair_default.return_value = mock_client
        mock_market_one = mock.Mock(
            market_id="1.234", update_market_catalogue=True, closed=False
        )
        mock_market_two = mock.Mock(
            market_id="5.678", update_market_catalogue=False, closed=False
        )
        mock_market_three = mock.Mock(
            market_id="9.101", update_market_catalogue=True, closed=True
        )
        mock_flumine.markets.markets = {
            "1.234": mock_market_one,
            "5.678": mock_market_two,
            "9.101": mock_market_three,
        }

        worker.poll_market_catalogue(mock_context, mock_flumine)
        mock_client.betting_client.betting.list_market_catalogue.assert_called_with(
            filter={"marketIds": ["1.234"]},
            market_projection=[
                "COMPETITION",
                "EVENT",
                "EVENT_TYPE",
                "RUNNER_DESCRIPTION",
                "RUNNER_METADATA",
                "MARKET_START_TIME",
                "MARKET_DESCRIPTION",
            ],
            max_results=25,
        )
        mock_flumine.handler_queue.put.assert_called_with(
            mock_events.MarketCatalogueEvent()
        )

    @mock.patch("flumine.worker.events")
    def test_poll_market_catalogue_status_error(self, mock_events):
        mock_context = mock.Mock()
        mock_flumine = mock.Mock()
        mock_client = mock.Mock()
        mock_flumine.clients.get_betfair_default.return_value = mock_client
        mock_market_one = mock.Mock(
            market_id="1.234", update_market_catalogue=True, closed=False
        )
        mock_flumine.markets.markets = {
            "1.234": mock_market_one,
        }
        mock_client.betting_client.betting.list_market_catalogue.side_effect = (
            exceptions.StatusCodeError("503")
        )

        worker.poll_market_catalogue(mock_context, mock_flumine)
        mock_client.betting_client.betting.list_market_catalogue.assert_called_with(
            filter={"marketIds": ["1.234"]},
            market_projection=[
                "COMPETITION",
                "EVENT",
                "EVENT_TYPE",
                "RUNNER_DESCRIPTION",
                "RUNNER_METADATA",
                "MARKET_START_TIME",
                "MARKET_DESCRIPTION",
            ],
            max_results=25,
        )
        mock_flumine.handler_queue.put.assert_not_called()

    @mock.patch("flumine.worker.events")
    def test_poll_market_catalogue_error(self, mock_events):
        mock_context = mock.Mock()
        mock_flumine = mock.Mock()
        mock_client = mock.Mock()
        mock_flumine.clients.get_betfair_default.return_value = mock_client
        mock_market_one = mock.Mock(
            market_id="1.234", update_market_catalogue=True, closed=False
        )
        mock_flumine.markets.markets = {
            "1.234": mock_market_one,
        }
        mock_client.betting_client.betting.list_market_catalogue.side_effect = (
            BetfairError()
        )

        worker.poll_market_catalogue(mock_context, mock_flumine)
        mock_client.betting_client.betting.list_market_catalogue.assert_called_with(
            filter={"marketIds": ["1.234"]},
            market_projection=[
                "COMPETITION",
                "EVENT",
                "EVENT_TYPE",
                "RUNNER_DESCRIPTION",
                "RUNNER_METADATA",
                "MARKET_START_TIME",
                "MARKET_DESCRIPTION",
            ],
            max_results=25,
        )
        mock_flumine.handler_queue.put.assert_not_called()

    @mock.patch("flumine.worker.events")
    def test_poll_account_balance(self, mock_events):
        mock_context = mock.Mock()
        mock_client = mock.Mock(account_funds={1: 2})
        mock_flumine = mock.Mock(clients=[mock_client])
        worker.poll_account_balance(mock_context, mock_flumine)
        mock_client.update_account_details.assert_called_with()
        mock_events.BalanceEvent.assert_called_with(mock_client)
        mock_flumine.log_control.assert_called_with(mock_events.BalanceEvent())

    @mock.patch("flumine.worker._get_cleared_market")
    @mock.patch("flumine.worker._get_cleared_orders")
    def test_poll_market_closure(
        self, mock__get_cleared_orders, mock__get_cleared_market
    ):
        mock_client = mock.Mock(
            id="123",
            paper_trade=False,
            market_recording_mode=False,
            EXCHANGE=ExchangeType.BETFAIR,
        )
        mock_client_sim = mock.Mock(
            id="456",
            paper_trade=False,
            market_recording_mode=False,
            EXCHANGE=ExchangeType.SIMULATED,
        )
        mock_flumine = mock.Mock(clients=[mock_client, mock_client_sim])
        market_one = mock.Mock(closed=False)
        market_two = mock.Mock(
            closed=True, orders_cleared=["123"], market_cleared=["123"]
        )
        market_three = mock.Mock(closed=True, orders_cleared=[], market_cleared=[])
        mock_flumine.markets.markets = {
            1: market_one,
            2: market_two,
            3: market_three,
        }
        worker.poll_market_closure({}, mock_flumine)
        mock__get_cleared_orders.assert_called_with(
            mock_flumine, mock_client.betting_client, market_three.market_id
        )
        mock__get_cleared_market.assert_called_with(
            mock_flumine, mock_client.betting_client, market_three.market_id
        )

    def test_poll_market_closure_paper(self):
        mock_client = mock.Mock(paper_trade=True)
        mock_flumine = mock.Mock(clients=[mock_client])
        mock_flumine.markets.markets = {1: mock.Mock()}
        worker.poll_market_closure({}, mock_flumine)

    @mock.patch("flumine.worker.config")
    @mock.patch("flumine.worker.events")
    def test__get_cleared_orders(self, mock_events, mock_config):
        mock_flumine = mock.Mock()
        mock_market = mock.Mock(blotter=[])
        mock_flumine.markets.markets = {
            "1.23": mock_market,
        }
        mock_betting_client = mock.Mock()
        mock_cleared_orders = mock.Mock()
        mock_cleared_orders.orders = []
        mock_cleared_orders.more_available = False
        mock_betting_client.betting.list_cleared_orders.return_value = (
            mock_cleared_orders
        )

        self.assertTrue(
            worker._get_cleared_orders(mock_flumine, mock_betting_client, "1.23")
        )
        mock_betting_client.betting.list_cleared_orders.assert_called_with(
            bet_status="SETTLED",
            market_ids=["1.23"],
            from_record=0,
            customer_strategy_refs=[mock_config.customer_strategy_ref],
        )
        mock_flumine.log_control.assert_called_with(mock_events.ClearedOrdersEvent())
        mock_flumine.handler_queue.put.assert_called_with(
            mock_events.ClearedOrdersEvent()
        )

    @mock.patch("flumine.worker.config")
    @mock.patch("flumine.worker.events")
    def test__get_cleared_orders_status_error(self, mock_events, mock_config):
        mock_flumine = mock.Mock()
        mock_betting_client = mock.Mock()
        mock_betting_client.betting.list_cleared_orders.side_effect = (
            exceptions.StatusCodeError("503")
        )
        self.assertFalse(
            worker._get_cleared_orders(mock_flumine, mock_betting_client, "1.23")
        )
        mock_betting_client.betting.list_cleared_orders.assert_called_with(
            bet_status="SETTLED",
            market_ids=["1.23"],
            from_record=0,
            customer_strategy_refs=[mock_config.customer_strategy_ref],
        )

    @mock.patch("flumine.worker.config")
    @mock.patch("flumine.worker.events")
    def test__get_cleared_orders_error(self, mock_events, mock_config):
        mock_flumine = mock.Mock()
        mock_betting_client = mock.Mock()
        mock_betting_client.betting.list_cleared_orders.side_effect = BetfairError
        self.assertFalse(
            worker._get_cleared_orders(mock_flumine, mock_betting_client, "1.23")
        )
        mock_betting_client.betting.list_cleared_orders.assert_called_with(
            bet_status="SETTLED",
            market_ids=["1.23"],
            from_record=0,
            customer_strategy_refs=[mock_config.customer_strategy_ref],
        )

    @mock.patch("flumine.worker.config")
    @mock.patch("flumine.worker.events")
    def test__get_cleared_orders_not_settled(self, mock_events, mock_config):
        mock_flumine = mock.Mock()
        mock_order = mock.Mock(size_matched=24.60)
        mock_market = mock.Mock(blotter=[mock_order])
        mock_flumine.markets.markets = {
            "1.23": mock_market,
        }
        mock_betting_client = mock.Mock()
        mock_cleared_orders = mock.Mock()
        mock_cleared_orders.orders = []
        mock_cleared_orders.more_available = False
        mock_betting_client.betting.list_cleared_orders.return_value = (
            mock_cleared_orders
        )

        self.assertFalse(
            worker._get_cleared_orders(mock_flumine, mock_betting_client, "1.23")
        )
        mock_betting_client.betting.list_cleared_orders.assert_called_with(
            bet_status="SETTLED",
            market_ids=["1.23"],
            from_record=0,
            customer_strategy_refs=[mock_config.customer_strategy_ref],
        )
        mock_flumine.log_control.assert_not_called()
        mock_flumine.handler_queue.put.assert_not_called()

    @mock.patch("flumine.worker.config")
    @mock.patch("flumine.worker.events")
    def test__get_cleared_market(self, mock_events, mock_config):
        mock_flumine = mock.Mock()
        mock_betting_client = mock.Mock()
        mock_cleared_markets = mock.Mock()
        mock_cleared_markets.orders = [1]
        mock_betting_client.betting.list_cleared_orders.return_value = (
            mock_cleared_markets
        )
        self.assertTrue(
            worker._get_cleared_market(mock_flumine, mock_betting_client, "1.23")
        )
        mock_betting_client.betting.list_cleared_orders.assert_called_with(
            bet_status="SETTLED",
            market_ids=["1.23"],
            group_by="MARKET",
            customer_strategy_refs=[mock_config.customer_strategy_ref],
        )
        mock_flumine.handler_queue.put.assert_called_with(
            mock_events.ClearedMarketsEvent()
        )

    @mock.patch("flumine.worker.config")
    def test__get_cleared_market_no_orders(self, mock_config):
        mock_flumine = mock.Mock()
        mock_betting_client = mock.Mock()
        mock_cleared_markets = mock.Mock()
        mock_cleared_markets.orders = []
        mock_betting_client.betting.list_cleared_orders.return_value = (
            mock_cleared_markets
        )
        self.assertFalse(
            worker._get_cleared_market(mock_flumine, mock_betting_client, "1.23")
        )
        mock_betting_client.betting.list_cleared_orders.assert_called_with(
            bet_status="SETTLED",
            market_ids=["1.23"],
            group_by="MARKET",
            customer_strategy_refs=[mock_config.customer_strategy_ref],
        )

    @mock.patch("flumine.worker.config")
    @mock.patch("flumine.worker.events")
    def test__get_cleared_market_status_error(self, mock_events, mock_config):
        mock_flumine = mock.Mock()
        mock_betting_client = mock.Mock()
        mock_betting_client.betting.list_cleared_orders.side_effect = (
            exceptions.StatusCodeError("503")
        )
        self.assertFalse(
            worker._get_cleared_market(mock_flumine, mock_betting_client, "1.23")
        )
        mock_betting_client.betting.list_cleared_orders.assert_called_with(
            bet_status="SETTLED",
            market_ids=["1.23"],
            group_by="MARKET",
            customer_strategy_refs=[mock_config.customer_strategy_ref],
        )

    @mock.patch("flumine.worker.config")
    @mock.patch("flumine.worker.events")
    def test__get_cleared_market_error(self, mock_events, mock_config):
        mock_flumine = mock.Mock()
        mock_betting_client = mock.Mock()
        mock_betting_client.betting.list_cleared_orders.side_effect = BetfairError
        self.assertFalse(
            worker._get_cleared_market(mock_flumine, mock_betting_client, "1.23")
        )
        mock_betting_client.betting.list_cleared_orders.assert_called_with(
            bet_status="SETTLED",
            market_ids=["1.23"],
            group_by="MARKET",
            customer_strategy_refs=[mock_config.customer_strategy_ref],
        )
